# -*- coding: utf-8 -*-
import sys
from Parser import Parser
from multiprocessing import Process, Queue, Pipe
import math
import time
import datetime
import commands
import struct
import socket

from math import atan2, cos, pi, sin, sqrt, tan

try:
    from gi.repository import Gtk 
    from gi.repository import GObject
except ImportError:
    sys.exit("pygtk not found.")


UI_FILE = "gui/tmy_monitor.glade"
PORT    = 9030
DATALEN = 2048

subsys = ['msap2', 'out', 'rgps', 'cm1', 'asu', 'tsu2', 'rcc', 'cm2', 'fsu', 
          'tsu1', 'msap1']

cols = {'Timestamp'  : None,
         'APIDQ'     : None,
         'PR'        : None,
         'USD'       : None,
         'IC4PS'     : None,
         'Timeout'   : None,
         'Online'    : None
         }

estado_subsys = {}

for syst in subsys:
    estado_subsys[syst] = cols.copy()

iface = {0 : 'msap2',
         1 : 'out',
         2 : 'rgps',
         3 : 'cm1',
         4 : 'asu',
         5 : 'tsu2',
         6 : 'rcc',
         7 : 'cm2',
         8 : 'fsu',
         9 : 'tsu1',
         10: 'msap1'
         }

# estado_subsys[iface[0]]['Timestamp']

# Estructura de Telemetria, parseo del string
# TLMY TLMY
N_ROUTERS = 11
tlmy_struct_iface = 'LLHBBIIIHfI'
tlmy_tlmy = '<LLLLLLHHLLfIBIffffHHBBIIHIIBBBB'
tlmy_struct = tlmy_tlmy + (N_ROUTERS * tlmy_struct_iface)

# Offsets estructura principal
starttime = 4
starttimeusec = 5
loadavg = 10
freemem = 11
mucount = 23
muerrcount = 24
gps_quality = 28
gps_sats = 29
obt_set = 30
curr_queue_packets = 27

# Offsets dentro de la estructura tlmy_iface
obt_input_info_iface      = 0
obt_input_info_iface_usec = 1
ifconf_iface              = 2
ifid_iface                = 3
apidq_iface               = 4
prc_iface                 = 5
usd_iface                 = 6
ic4ps_iface               = 7
prf_iface                 = 8
ifdr_iface                = 9
timeout_iface             = 10

# RGPS 0x0C0 (msj70)
rgpssolnav = '>BiiiiiiBBBHIBIIHHHHBBBBBB'
ecefx = 1
ecefy = 2
ecefz = 3
velx  = 4
vely  = 5
velz  = 6
week  = 10
tow   = 11
svs   = 12

# RGPS 0x0C4 (msj61) nextppstimetag
nextgpstimetag = '>BBIB'
pfield  = 1
tfield  = 2
quality = 3

def ecef2geodetic(x, y, z, degrees=True):
    # Constants (WGS ellipsoid)
    a = 6378137
    e = 8.1819190842622e-2
    # Calculation
    b = sqrt(pow(a,2) * (1-pow(e,2)))
    ep = sqrt((pow(a,2)-pow(b,2))/pow(b,2))
    p = sqrt(pow(x,2)+pow(y,2))
    th = atan2(a*z, b*p)
    lon = atan2(y, x)
    lat = atan2((z+ep*ep*b*pow(sin(th),3)), (p-e*e*a*pow(cos(th),3)))
    n = a/sqrt(1-e*e*pow(sin(lat),2))
    alt = p/cos(lat)-n
    lat = (lat*180)/pi
    lon = (lon*180)/pi
    return (lat, lon, alt)

class TMY_MONITOR_WINDOW(Gtk.Window):
    def __init__(self):
        Gtk.Window.__init__(self)

        # Cosas de la GUI
        self.builder     = Gtk.Builder()
        self.builder.add_from_file(UI_FILE)
        
        # Ventana principal
        self.window      = self.builder.get_object('tlmyStatus')
        
        # Crear tabla de subsistemas
        self.subsys      = self.create_table()

        # Barra de estado
        self.statusbar  = self.builder.get_object('statusbarTMY')
        self.context_id = self.statusbar.get_context_id('progress_message')

        # Informacion Frame CCSDS
        self.frameArrival = self.builder.get_object('frameArrival')
        self.frameNr      = self.builder.get_object('frameNr')
        self.CCSDSPkts    = self.builder.get_object('CCSDSPkts')
        self.APIDs        = self.builder.get_object('APIDS')
        self.framessec    = self.builder.get_object('framessec')
        self.datarate     = self.builder.get_object('datarate')
        self.packetssec   = self.builder.get_object('packetssec')

        # Estampas de tiempo
        self.unixfromtfield  = self.builder.get_object('unixfromtfield')
        self.civilfromtfield = self.builder.get_object('civilfromtfield')
        self.unixlocalhost   = self.builder.get_object('unixlocalhost')
        self.civillocalhost  = self.builder.get_object('civillocalhost')

        # Estado Sistema
        self.sysload      = self.builder.get_object('load')
        self.freemem      = self.builder.get_object('freemem')
        self.muwrites     = self.builder.get_object('muwrites')
        self.muerrwrites  = self.builder.get_object('muerrwrites')
        self.curmsg       = self.builder.get_object('curmsg')

        # RGPS
        self.rgpssol      = self.builder.get_object('rgpssol')
        self.rgpsats      = self.builder.get_object('rgpssats')
        self.obtstatus    = self.builder.get_object('obtstatus')
        self.pfield       = self.builder.get_object('pfield')
        self.tfield       = self.builder.get_object('tfield')
        self.quality      = self.builder.get_object('quality')

        self.lat          = self.builder.get_object('lat')
        self.lon          = self.builder.get_object('lon')
        self.alt          = self.builder.get_object('alt')
        
        self.vel          = self.builder.get_object('vel')
        self.week         = self.builder.get_object('weekNr')
        self.tow          = self.builder.get_object('tow')

        # Informacion Telemetria
        self.ccsdsPK      = self.builder.get_object('ccsdsPK')
        self.timestamp    = self.builder.get_object('timestamp')
        self.starttime    = self.builder.get_object('starttime')
        self.uptime       = self.builder.get_object('uptime')

        # Control
        self.filechooser = self.builder.get_object('filechooser')
        self.archivoxapid = self.builder.get_object('archivoxapid')

        # Spinner
        self.spinner     = self.builder.get_object('spinner')

        # Conectar señales
        self.builder.connect_signals(self)

        # Control, cosas que no son de la GUI
        self.running           = False
        self.parser            = None
        self.queue             = None
        self.this              = None
        self.parserProcessPipe = None
        self.framecounter      = 0
        self.packetcounter     = 0
        self.record            = False
        self.archivoxapid.set_active(False)

        # Seteos de estados iniciales
        self.rgpssol.set_from_stock   ('gtk-no', 0)
        self.obtstatus.set_from_stock ('gtk-no', 0)
        self.statusbar.push(self.context_id, 'Sistema Detenido')
        self.outpath = self.filechooser.get_current_folder_uri()[7:]
        self.datatime  = None
        self.outframes = None
        self.outapids  = None

        for subsys in estado_subsys.keys():
            estado_subsys[subsys]['Online'].set_from_stock('gtk-no', 0)

    def start(self):
        # Mostramos la ventana
        
        self.window.show_all()
        # Llamada a GTK main
        Gtk.main()

    def stop(self):
        self.on_stop_clicked(None)
        Gtk.main_quit()
        sys.exit()


    def create_table(self):
        global estado_subsys
        for key in estado_subsys.keys():
            for otherkey in estado_subsys[key].keys():
                estado_subsys[key][otherkey] = self.builder.get_object(key+otherkey)


    def create_files(self):
        self.outpath = self.filechooser.get_current_folder_uri()[7:]
        self.datatime = datetime.datetime.now()
        d = self.datatime
        self.outframes = open(self.outpath+'/data-%d%02d%02d%02d%02d%02d.dat' % (d.year, 
                                                                                 d.month, 
                                                                                 d.day, 
                                                                                 d.hour, 
                                                                                 d.minute, 
                                                                                 d.second), 
                              'wb')

        
        if self.archivoxapid.get_active():
            self.outapids = {}
        
        return True

    def close_files(self):
        if self.outframes:
            self.outframes.close()
        if self.archivoxapid.get_active():
            for key in self.outapids.keys():
                self.outapids[key].close()
            
        return True

    #### Eventos y Callbacks
    # Ventana
    def on_tlmyStatus_destroy (self, widget):
        self.stop()

    # Control
    def on_start_clicked (self, widget):
        if not self.running:
            self.spinner.start()
            self.running = True
            self.framecounter = 0
            self.packetcounter = 0
            GObject.timeout_add(1000, self.calcframes)
            self.statusbar.push(self.context_id, 'Sistema Activo')
            if self.record:
                self.create_files()
            self.process()
        else:
            message = Gtk.MessageDialog(self, 0, Gtk.MessageType.ERROR,
                                        Gtk.ButtonsType.OK, 
                                        'El sistema ya se encuentra corriendo')
            message.run()
            message.destroy()
        

    def on_stop_clicked (self, widget):
        if self.running:
            self.rgpssol.set_from_stock  ('gtk-no', 0)
            self.rgpsats.set_text        ('-')
            self.obtstatus.set_from_stock('gtk-no', 0)
            self.spinner.stop()
            self.running = False
            self.filechooser.set_sensitive(True)
            self.archivoxapid.set_sensitive(True)
            if self.record:
                self.close_files()
            self.statusbar.push(self.context_id, 'Sistema Detenido')

    def on_record_toggled (self, widget):
        if self.record:
            self.record = False
            self.archivoxapid.set_sensitive(True)
            self.filechooser.set_sensitive(True)
            if self.running:
                self.close_files()
        else:
            self.archivoxapid.set_sensitive(False)
            self.filechooser.set_sensitive(False)
            self.record = True
            if self.running:
                self.create_files()
            
    # Menu
    # Archivo
    def on_menuQuit_activate (self, widget):
        self.stop()
    
    # Salvar
    def on_menuSave_activate (self, widget):
        pass

    # Editar
    def on_menuPropierties_activate (self, widget):
        pass

    # Ayuda
    def on_menuAbout_activate (self, widget):
        about = Gtk.AboutDialog()
        about.set_program_name("VEX Monitor de Sistema de Telemetria")
        about.set_version("1.0")
        about.set_copyright("Copyright Fernando P. Hauscarriaga <fernandoph@iar.unlp.edu.ar>\nInstituto Argentino de Radioastronomía (IAR)\nVehiculos Espaciales de la Nueva Generación S.A. (VENG SA)\nComisión Nacional de Actividades Espaciales (CONAE)\nMinisterio de Planificación Federal, Inversion Pública y Servicios\nPresidencia de la Nación\nRepública Argentina")
        about.set_authors(["Fernando P. Hauscarriaga"])
        about.set_website("http://www.iar.unlp.edu.ar")
        about.set_website_label("Instituto Argentino de Radioastronomía")

        about.run()
        about.destroy()

    # Procesamiento de los frames que llegan
    def process(self):
        server_address = (('', PORT))
        s = socket.socket(socket.AF_INET, socket.SOCK_DGRAM)
        s.settimeout(0.1)
        s.bind(server_address)
 
        while self.running:
            frame = Parser.TransferFrame()
            try:
                # Adquirir frame
                data  = s.recv(2048)
                # Parseo
                frame.parseFrame(data)
                # Actualizacion barra de estados
                self.statusbar.push(self.context_id, 'Adquiriendo frame %d' % frame.getMasterChannelFrameCount())

                # Grabacion
                if self.record:
                    self.outframes.write(data)
                    if self.archivoxapid.get_active():
                        for packet in frame.getPackets():
                            if not self.outapids.has_key(packet.getAPID()):
                                d = self.datatime
                                self.outapids[packet.getAPID()] = open(self.outpath+'/data_apid_%03X-%d%02d%02d%02d%02d%02d.dat' % (packet.getAPID(), d.year, d.month, d.day, d.hour, d.minute, d.second), 'wb')
                            self.outapids[packet.getAPID()].write(packet.getRawPacket())
                            
                # Actualizar toda la gui con la info del frame
                self.updateGUI(frame)
            except:
                self.statusbar.push(self.context_id, 'Timeout Recibiendo Paquetes')

            while Gtk.events_pending():
                Gtk.main_iteration()

    # Calculo frames x segundo y datarate
    def calcframes (self):
        if self.running:
            self.framessec.set_text(str(self.framecounter))
            self.datarate.set_text('%d KB/Seg' % ((self.framecounter*2048)/1024))
            self.packetssec.set_text(str(self.packetcounter))
            self.framecounter = 0
            self.packetcounter = 0
            return True
        else:
            return False
        
    # Actualizar la GUI
    def updateGUI (self, frame):
        toa = frame.getToa().time()
        self.frameArrival.set_text(str(toa))
        self.frameNr.set_text(str(frame.getMasterChannelFrameCount()))
        pkts  = ''
        count = 0
        self.framecounter += 1
        for pk in frame.getPackets():
            pkts += str('0x%02X ' % pk.getAPID())
            count += 1
        self.CCSDSPkts.set_text(str(count))
        self.APIDs.set_text(pkts)
        
        for packet in frame.getPackets():
            self.packetcounter += 1
            if packet.getAPID() == 0x120:
                data = struct.unpack(tlmy_struct, packet.getRawPacket()[13:])
                self.sysload.set_text(str('%.2f' % data[loadavg]))
                self.freemem.set_text(str(data[freemem]/1024)+' Kb')
                self.muwrites.set_text(str(data[mucount]))
                self.muerrwrites.set_text(str(data[muerrcount]))
                self.curmsg.set_text(str(data[curr_queue_packets]))

                if data[gps_quality] == 1:
                    self.rgpssol.set_from_stock('gtk-yes', 0)
                else:
                    self.rgpssol.set_from_stock('gtk-no',  0)

                self.rgpsats.set_text(str(data[gps_sats]))

                if data[obt_set] == 1:
                    self.obtstatus.set_from_stock('gtk-yes', 0)
                else:
                    self.obtstatus.set_from_stock('gtk-no',  0)

                self.ccsdsPK.set_text(str(packet.getPacketSequenceCount()))
                self.timestamp.set_text(str('%ld.%06ld' % (data[0], data[1])))
                self.starttime.set_text(str('%ld.%06ld' % (data[starttime], data[starttimeusec])))
                
                t0 = datetime.datetime.fromtimestamp(data[starttime] + data[starttimeusec] * 10 ** -6)
                t1 = datetime.datetime.fromtimestamp(data[0] + data[1] * 10 ** -6)
                self.uptime.set_text(str(t1-t0))

                # Actualizacion de la tabla de subsistemas
                for n in range (0, N_ROUTERS):
                    if data[31 + n*(len(tlmy_struct_iface)) + ifid_iface] == n:
                        estado_subsys[iface[n]]['Online'].set_from_stock ('gtk-yes', 0)
                    else:
                        estado_subsys[iface[n]]['Online'].set_from_stock ('gtk-no', 0)
                        

                for n in range (0, N_ROUTERS):
                    estado_subsys[iface[n]]['Timestamp'].set_text(str(data[31 + n*(len(tlmy_struct_iface)) + obt_input_info_iface]))
                    estado_subsys[iface[n]]['APIDQ'].set_text(str(data[31 + n*(len(tlmy_struct_iface)) + apidq_iface ]))
                    estado_subsys[iface[n]]['PR'].set_text(str(data[31 + n*(len(tlmy_struct_iface)) + prc_iface ]))
                    estado_subsys[iface[n]]['USD'].set_text(str(data[31 + n*(len(tlmy_struct_iface)) + usd_iface ]))
                    estado_subsys[iface[n]]['IC4PS'].set_text(str( data[31 + n*(len(tlmy_struct_iface)) + ic4ps_iface ]))
                    estado_subsys[iface[n]]['Timeout'].set_text(str(data[31 + n*(len(tlmy_struct_iface)) + timeout_iface]))

            if packet.getAPID() == 0x0C0:
                data = struct.unpack(rgpssolnav, packet.getRawPacket()[13:])
		print data[ecefx], data[ecefy], data[ecefz]
                geo = ecef2geodetic(data[ecefx], data[ecefy], data[ecefz])
                self.lat.set_text(str(geo[0]))
                self.lon.set_text(str(geo[1]))
                self.alt.set_text(str(geo[2]))
                self.vel.set_text(str(sqrt(data[velx]**2 + data[vely]**2 + data[velz]**2)))
                self.week.set_text(str(data[week]))
                self.tow.set_text(str(data[tow]))
                
            if packet.getAPID() == 0x0C4:
                data = struct.unpack(nextgpstimetag, packet.getRawPacket()[13:])
                self.pfield.set_text('0x%02x' % data[pfield])
                self.tfield.set_text(str(data[tfield]))
                self.quality.set_text(str(data[quality]))

                # Actualizar estampas de tiempo
                self.unixfromtfield.set_text(str(data[tfield] + 315964800 - 16 - 1))
                self.civilfromtfield.set_text(str(time.ctime(data[tfield] + 315964800 - 16 - 1)))
                self.unixlocalhost.set_text(str(time.time() + 10800))
                self.civillocalhost.set_text(str(time.ctime(time.time() + 10800)))

def main():
    gui    = TMY_MONITOR_WINDOW()
    gui.start()

if __name__ == "__main__":
    main()
