import struct
import sys
import socket
import datetime
import pickle

#from multiprocessing import Process, Queue
import datetime

# Sincronizador :-)
def receive_tmy_frames(q, packetsize, server_address):
    pattern = struct.pack ("BBBB", 0x1A, 0xCF, 0xFC, 0x1D)
    s = socket.socket(socket.AF_INET, socket.SOCK_DGRAM)
    s.bind(server_address)
    buf = ''

    while True:
        while True:
            buf         = buf + s.recv(packetsize)
            startoffset = buf.find(pattern)
            if startoffset >= 0:
                while True:
                    buf = buf + s.recv(packetsize)
                    stopoffset = buf.find(pattern, startoffset+1)
                    if stopoffset >= 0:
                        break
            else:
                continue
            break

        q.put(buf[startoffset:stopoffset])
        buf = buf[stopoffset:]

###############################################################################
## 
## Clase SpacePacket                   
##
###############################################################################
class SpacePacket():
    def __init__(self):
        self.packetVersionNumber = None
        self.packetType          = None
        self.secHdrFlag          = None
        self.APID                = None
        self.sequenceFlags       = None
        self.packetSequenceCount = None
        self.packetDataLength    = None
        self.pksechdrpfield      = None
        self.pksechdrtfieldc     = None
        self.pksechdrtfieldf     = None
        self.payload             = None
        self.rawpacket           = None

    def setPacketVersionNumber(self, aVersion):
        self.packetVersionNumber = aVersion

    def setPacketType(self, aType):
        self.packetType = aType

    def setSecHdrFlag(self, aSecHdrFlag):
        self.secHdrFlag = aSecHdrFlag

    def setAPID (self, anAPID):
        self.APID = anAPID

    def setSequenceFlags(self, aSequenceFlag):
        self.sequenceFlags = aSequenceFlag

    def setPacketSequenceCount(self, aCount):
        self.packetSequenceCount = aCount

    def setPacketDataLength(self, aLength):
        self.packetDataLength = aLength
        
    def setPkSecHdrPfield (self, anOctect):
        self.pksechdrpfield = anOctect

    def setPkSecHdrTfieldC (self, aTfieldc):
        self.pksechdrtfieldc = aTfieldc

    def setPkSecHdrTfieldF (self, aTfieldf):
        self.pksechdrtfieldf = aTfieldf

    def setPayload(self, aPayload):
        self.payload = aPayload

    def setRawPacket(self, data):
        self.rawpacket = data

    def getPacketVersionNumber(self):
        return self.packetVersionNumber

    def getPacketType(self):
        return self.packetType

    def getSecHdrFlag(self):
        return self.secHdrFlag

    def getAPID(self):
        return self.APID

    def getSequenceFlags(self):
        return self.sequenceFlags

    def getPacketSequenceCount(self):
        return self.packetSequenceCount

    def getPacketDataLength(self):
        return self.packetDataLength

    def getSecHdrPField(self):
        return self.pksechdrpfield

    def getSecHdrTFieldC(self):
        return self.pksechdrtfieldc

    def getSecHdrTFieldF(self):
        return self.pksechdrtfieldf

    def getPayload(self):
        return self.payload

    def getRawPacket(self):
        return self.rawpacket

    # Operaciones
    def parseHeader(self, data):
        self.setAPID               (struct.unpack('>H', data[0:2])[0] & 0x07FF)
        self.setPacketSequenceCount(struct.unpack('>H', data[2:4])[0] & 0x3FFF)
        self.setPacketDataLength   (struct.unpack('>H', data[4:6])[0])

        if struct.unpack('>H', data[0:2])[0] & 0x800:
            self.setSecHdrFlag(1)

        #print struct.unpack('>H', data[2:4])[0] & 0x3FFF

    def parseSecHdr(self, data):
        self.setPkSecHdrPfield (struct.unpack('B', data[0])[0])
        self.setPkSecHdrTfieldC(struct.unpack('>L', data[1:5])[0])
        self.setPkSecHdrTfieldF(struct.unpack('>H', data[5:7])[0])

    def isIDLE(self):
        if self.getAPID() == 0x07FF:
            return True
        else:
            return False
    
    def savePacketPayload(self, data, pos, pklen):
        self.setPayload(data[pos:pos+pklen+1])


##############################################################################
## 
## Clase TransferFrame           
##
##############################################################################  
class TransferFrame():
    def __init__(self):
        # Cabecera
        self.asm                          = None
        self.spacecraftId                 = None
        self.virtualChannelId             = None
        self.OCFFlag                      = None
        self.masterChannelFrameCount      = None
        self.virtualChannelFrameCount     = None
        self.transferFrameDataFieldStatus = None
        # Payload
        self.payload                      = None
        
        # Frame crudo, binario
        self.binFrame = ''

        self.packets = []
        self.toa     = 0

        # Constantes
        self.ASM = 0x1ACFFC1D # Bytes de sincronismo

    # Setters
    def setASM (self, asm):
        self.asm = asm

    def setSpacecraftId(self, anId):
        self.spacecraftId = anId

    def setVirtualChannelId(self, anId):
        self.virtualChannelId = anId

    def setOCFFlag(self, aFlag):
        self.OCFFlag = aFlag

    def setMasterChannelFrameCount(self, aNumber):
        self.masterChannelFrameCount = aNumber

    def setVirtualChannelFrameCount(self, aNumber):
        self.virtualChannelFrameCount = aNumber

    def setTransferFrameDataFieldStatus(self, aFieldStatus):
        self.transferFrameDataFieldStatus = aFieldStatus

    def setPayload(self, aPayload):
        self.payload = aPayload

    def setBinFrame(self, aBinaryString):
        self.binFrame = aBinaryString

    def setPackets(self, aListOfPackets):
        self.packets = aListOfPackets

    def setToa(self, aTimeOfArrival):
        self.toa = aTimeOfArrival

    # Getters
    def getASM (self):
        return self.asm

    def getSpacecraftId(self):
        return self.spacecraftId

    def getVirtualChannelId(self):
        return self.virtualChannelId

    def getOCFFlag(self):
        return self.OCFFlag

    def getMasterChannelFrameCount(self):
        return self.masterChannelFrameCount

    def getVirtualChannelFrameCount(self):
        return self.virtualChannelFrameCount

    def getTransferFrameDataFieldStatus(self):
        return self.transferFrameDataFieldStatus

    def getPayload(self):
        return self.payload
        
    def getBinFrame(self):
        return self.binFrame

    def getPackets(self):
        return self.packets

    def getToa(self):
        return self.toa

    # Operaciones
    def appendData(self, stringData):
        self.binFrame = self.binFrame + stringData

    def parseFrame(self, data):
        #00 0E 00 00 1F FE --> Es el header de un frame idle del tmy adapter
        idleframe = struct.pack ('BBBBBB', 0x00, 0x0E, 0x00, 0x00, 0x1F, 0xFE)

        if data[4:10].find(idleframe) < 0:
            self.appendData(data)
            self.toa = datetime.datetime.now()

        #for byte in data[0:11]:
        #    print '0x%02X' % ord(byte),
        #print

            self.setMasterChannelFrameCount (struct.unpack('=B', data[6])[0])
            self.setVirtualChannelFrameCount(struct.unpack('=B', data[7])[0])
            self.setPayload(data[10:])

            pos = 0
            
            while True:
                pk = SpacePacket()
                pk.parseHeader(self.getPayload()[pos:pos+6])

                if pk.getSecHdrFlag() == 1:
                    pk.parseSecHdr(self.getPayload()[pos+6:pos+13])
                    
                if pk.getSecHdrFlag() == 1:
                    pk.setRawPacket(self.getPayload()
                                    [pos:pos+pk.getPacketDataLength()+7])
                else:
                    pk.setRawPacket(self.getPayload()
                                    [pos:pos+pk.getPacketDataLength()])

                pos += 6

                pk.savePacketPayload(self.getPayload(), pos, 
                                     pk.getPacketDataLength())

                pos = pos + pk.getPacketDataLength() + 1

                self.packets.append(pk)

                if pos >= 2025 or pk.getAPID() == 0x7FF:  #
                    break

##############################################################################
##############################################################################
##############################################################################
#                                PARSERS
##############################################################################
## 
## Clase VexParser
##
##############################################################################
class VexParser():
    def __init__(self, aMessageQueue = None):
        self.file        = False
        self.parsedFile  = False
        self.queue       = aMessageQueue

        if self.queue == None:
            self.queue = Queue()

    # Setters
    def setQueue(self, aMessageQueue):
        self.queue = aMessageQueue

    def saveToFile(self):
        self.file = True
        now = datetime.datetime.now()
        self.file = open('vex_tmy-%4d%02d%02d%02d%02d%02d.dat' % 
                         (now.year,
                          now.month,
                          now.day,
                          now.hour,
                          now.minute,
                          now.second), 'wb')
        
        self.parsedfile = open ('vex_tmy-%4d%02d%02d%02d%02d%02d-parsed.dat' % 
                                (now.year, 
                                 now.month, 
                                 now.day, 
                                 now.hour, 
                                 now.minute, 
                                 now.second), 'wb')

    # Getters
    def getQueue(self):
        return self.queue

    # Funcionalidad

    # ASM
    def syncMe(self):
        # 0x1ACFFC1D
        self.waitByte(0x1A)
        self.waitByte(0xCF)
        self.waitByte(0xFC)
        self.waitByte(0x1D)

##############################################################################
## 
## Clase VexLiveParser(VexParser)
##
##############################################################################
# El ingestor transmite de a 250 bytes
class VexLiveParser(VexParser):
    def __init__(self, anIP = None, aPort = None, aDatalen = 2048):
        VexParser.__init__(self)

        self.socket  = None
        self.ip      = anIP
        self.port    = aPort
        self.datalen = aDatalen

        self.running = False

    def setIp(self, anIpAddress):
        self.ip = anIPAddress

    def setPort(self, aPort):
        self.port = aPort

    def getIp(self):
        return self.ip

    def getPort(self):
        return self.port

    def stopParse(self):
        print 'STOP'
        self.running = False

    def startParse(self, apipe):
        self.running = True

        qsocket = Queue()
        server_address = ((self.ip, self.port))

        # Si conectamos directamente al UDP de TLMY no hace falta sincronizar
        if self.datalen == 2048:
            s = socket.socket(socket.AF_INET, socket.SOCK_DGRAM)
            s.bind(('', 9030))
        # Si queremos conectar al ingestor, entonces hay que sincronizar
        else:
            recvProcess = Process(target=receive_tmy_frames, 
                                   args=(qsocket, self.datalen, server_address))
            recvProcess.start()

        if self.file:
            now = datetime.datetime.now()
            self.file = open('vex_tmy-%4d%02d%02d%02d%02d%02d.dat' % 
                             (now.year,
                              now.month,
                              now.day,
                              now.hour,
                              now.minute,
                              now.second), 'wb')
            
            self.parsedfile = open ('vex_tmy-%4d%02d%02d%02d%02d%02d-parsed.dat' % (now.year, now.month, now.day, now.hour, now.minute, now.second), 'wb')

        while self.running:
            try:
                curFrame = TransferFrame()
                #curFrame.parseFrame(qsocket.get(timeout = 0.00001))
                #curFrame.parseFrame(qsocket.get())
                if self.datalen == 2048:
                    curFrame.parseFrame(s.recv(2048))
                else:
                    curFrame.parseFrame(qsocket.get_nowait())
       
                if curFrame.getMasterChannelFrameCount() != None:
                    self.queue.put(curFrame)

                    if self.file:
                        self.file.write(curFrame.getBinFrame())
                        pickle.dump(curFrame, self.parsedfile)
            except:
                continue

            if apipe.poll():
                self.running = False
            
        recvProcess.terminate()
 
##############################################################################
## 
## Clase VexFileParser(VexParser)
##
##############################################################################
class VexFileParser(VexParser):
    def __init__(self, anInFile):
        VexParser.__init__(self)
        self.infile = anInFile
        
    def setinfile(self, anInFile):
        self.infile = anInFile

    def getinfile(self):
        return self.infile
        
    # Funcionalidad
    def startParse(self):
        if self.infile == None:
            return None
        
        fd = open(self.infile, 'rb')

        while (fd.read(1) != ''):
            fd.seek(fd.tell() - 1)
            curFrame = TransferFrame()
            curFrame.parseFrame(fd.read(2048))
            
            if curFrame.getMasterChannelFrameCount() != None:
                self.queue.put(curFrame)

        fd.close()
            
            
            
