#!/usr/bin/python

import socket, struct, time, sys, os
from optparse import OptionParser

from Parser import TransferFrame, SpacePacket

apid2txt        = {}
apid2txt[0x0A0] = 'ngct1'
apid2txt[0x0B0] = 'ngct2'
apid2txt[0x0C0] = 'rgpsextnav'
apid2txt[0x0C1] = 'rgpssats'
apid2txt[0x0C2] = 'rgpschinfo'
apid2txt[0x0C3] = 'rgpsinfomon'
apid2txt[0x0C4] = 'nextppstimetag'
apid2txt[0x0E0] = 'rcc'
apid2txt[0x0F0] = 'pfssup'
apid2txt[0x0F1] = 'pfsinf'
apid2txt[0x100] = 'amubatt'
apid2txt[0x110] = 'tvcsup'
apid2txt[0x111] = 'tvcinf'
apid2txt[0x090] = 'FSU'
apid2txt[0x120] = 'tlmy'
apid2txt[0x7FF] = 'idle'

iface2txt       = {}
iface2txt[0]  = 'msap2'
iface2txt[1]  = 'out'
iface2txt[2]  = 'rgps'
iface2txt[3]  = 'cm1'
iface2txt[4]  = 'asu'
iface2txt[5]  = 'tsu2'
iface2txt[6]  = 'rcc'
iface2txt[7]  = 'cm2'
iface2txt[8]  = 'fsu'
iface2txt[9]  = 'tsu1'
iface2txt[10] = 'msap1'

# Parseo de datos dentro de los space packets

# TLMY TLMY
N_ROUTERS = 11
tlmy_struct_iface = 'LLHBBIIIHfI'
tlmy_tlmy = '<LLLLLLHHLLfIBIffffHHBBIIHIIBBBB'
tlmy_struct = tlmy_tlmy + (N_ROUTERS * tlmy_struct_iface)

def main():
    fdout = {}

    parsecmdline = OptionParser()
    parsecmdline.add_option('-i', '--filename',
                            action='store', type='string', dest='filename',
                            help='Archivo de entrada')
    parsecmdline.add_option('-p', '--packet', action='store_true', 
                            dest='packet', help='Usar modo packet')
    parsecmdline.add_option('-f', '--frame', action='store_true', 
                            dest='frame', help='Usar modo frame')
    parsecmdline.add_option('-n', '--counters', action='store_true', 
                            dest='counters', 
                            help='Mostrar contadores (primero y ultimo)')
    parsecmdline.add_option('-s', '--startchop',
                            action='store', type='int', dest='startchop',
                            help='Cortar desde')
    parsecmdline.add_option('-t', '--tlmy',action='store_true', 
                            dest='tlmy', help='Generar tabla con datos de tlmy tlmy')
    parsecmdline.add_option('-v', '--verify',
                            action='store', type='int', dest='valverif',
                            default = -1,
                            help='Comparar cada byte del paquete con este byte')
    parsecmdline.add_option('-o', '--outfile',
                            action='store', type='string', dest='outfile',
                            help='Archivo de salida')

    (options, args) = parsecmdline.parse_args()

    if len(sys.argv) <= 1:
        print 'ERROR: Debe especificar argumentos.'
        print
        parsecmdline.print_help()
        return -1

    if options.filename == None:
        print 'ERROR: Especifique archivo de entrada.'
        print
        parsecmdline.print_help()
        return -1

    if options.frame and options.packet:
        print 'ERROR: debe elegir modo packet o modo frame, no ambos.'
        print
        parsecmdline.print_help()
        return -1

    if (options.frame or options.packet) and (options.counters == None  and
                                              options.tlmy     == None and
                                              options.stopchop == None  and
                                              options.valverif == -1):
        print 'ERROR: Especifique la operacion'
        print
        parsecmdline.print_help()
        return -1

    if options.frame and options.valverif >= 0:
        print 'ERROR: Argumentos incompatibles'
        print
        parsecmdline.print_help()
        return -1

    fd = open(options.filename, 'rb')

    if options.frame:
        fdout  = open(options.filename+'-Reporte_Frames.txt', 'w')
        fdout2 = open(options.filename+'-Tabla_Paquetes.txt', 'w')
    if options.tlmy:
        fdout3 = open(options.filename+'-TLMY-TLMY.txt', 'w')

    # Variables para control y estadistica
    framecount  = 0
    packetcount = 0
    firstframe  = None
    lastframe   = None
    firstpacket = None
    lastpacket  = None
    frameloss   = 0
    frameant    = 0
    packetant   = 0
    packetloss  = 0
    totvalverif = 0
    totbytes    = 0

    apidcount     = {}
    apidsizes     = {}
    badapidoffset = {}
    tapid         = {}

    valverif      = {}

    pksechdrcount = 0

    tpackets = []

    avgload  = []

    firstmem = 0
    lastmem  = 0

    if options.tlmy:
        tlmytlmy = []
    
    ##

    print '==============================================================================='
    print 'Abriendo: %s' % options.filename
    print 'Procesando, espere...'
    print
    
    while fd.read(1) != '':
        fd.seek(fd.tell() - 1)

        #### Operaciones con Frames
        if options.frame:
            curFrame = TransferFrame()
            curFrame.parseFrame(fd.read(2048))

            framecount = framecount + 1

            if firstframe == None:
                firstframe = (framecount, 
                              curFrame.getMasterChannelFrameCount())
                frameant   = curFrame.getMasterChannelFrameCount()

            lastframe = (framecount, curFrame.getMasterChannelFrameCount())

            # Calculo de frames perdidos y reporte
            #if frameant != 0 and firstframe != None:
            #    diff = curFrame.getMasterChannelFrameCount() - frameant
            #    if diff != 1: #and diff > 0:
            #        frameloss = frameloss + diff - 1
            #        print 'WARN: Frame perdido - ant.: %d, actual: %d en 0x%02X' % (frameant, curFrame.getMasterChannelFrameCount(), fd.tell())
            if firstframe != None:
                diff = curFrame.getMasterChannelFrameCount() - frameant

                if diff <> 1 and diff <> -255:
                    if diff < 0 and frameant <> 255:
                        diff = 255 + diff
                        #diff = 255 - frameant
                    frameloss += diff
                    print 'WARN: Frame perdido - ant.: %d, actual: %d en 0x%02X' % (frameant, curFrame.getMasterChannelFrameCount(), fd.tell())


            frameant = curFrame.getMasterChannelFrameCount()
            
            # Operaciones para los paquetes dentro del frame
            for pk in curFrame.getPackets():
                # Guardo APIDs que se encontraron
                if not apidcount.has_key(pk.getAPID()):
                    apidcount[pk.getAPID()] = 0

                if not apidsizes.has_key(pk.getAPID()):
                    apidsizes[pk.getAPID()] = 0

                if not apid2txt.has_key(pk.getAPID()):
                    #badapidoffset[pk.getAPID()] = fd.tell() - 2048
                    badapidoffset[pk.getAPID()] = 0

                if not apid2txt.has_key(pk.getAPID()) and badapidoffset[pk.getAPID()] == 0:
                    badapidoffset[pk.getAPID()] += 1

                if pk.getSecHdrFlag() == 1:
                    if not tapid.has_key(pk.getAPID()):
                        tapid[pk.getAPID()] = [0, []]
                    
                # Contador de APIDs    
                apidcount[pk.getAPID()] = apidcount[pk.getAPID()] + 1
                # Contador de APIDs no declarados
                apidsizes[pk.getAPID()] = pk.getPacketDataLength() + 1

                if pk.getSecHdrFlag() == 1:
                    tapid[pk.getAPID()][0] += 1;
                    t = pk.getSecHdrTFieldC() + (pk.getSecHdrTFieldF() * .001)
                    tapid[pk.getAPID()][1].append(t)

                if pk.getSecHdrFlag():
                    pksechdrcount += 1

                # Hacer una tabla con los paquetes
                # Formato:
                # # CCSDS Frame, # CCSDS Packet, Timestamp (OBT format), APID
                if pk.getSecHdrFlag():
                    tpackets.append('%d %d %.03f 0x%03X\n' % (curFrame.getMasterChannelFrameCount(),
                                              pk.getPacketSequenceCount(),
                                              pk.getSecHdrTFieldC() + pk.getSecHdrTFieldF()*(10**-3),
                                              pk.getAPID()))

                # Hacer una tabla con los valores de tlmy tlmy
                # Formato:
                # Timestamp(Formato OBT), # CCSDS TLMY, SysLoad, MemFree, MuTxCount, MuTxErr, QueuePKTS
                if options.tlmy:
                    if pk.getAPID() == 0x120:
                        tmp = struct.unpack(tlmy_struct, pk.getRawPacket()[13:])
                        s = ''
			# Formato tabla routers
			# Timestamp (unix) ifid apidq prc usd ic4ps timeout
                        for n in range(0, N_ROUTERS):
                            s += '%d %d %d %d %d %d %d ' % (tmp[31 + n*(len(tlmy_struct_iface)) + 0 ],
                                                            tmp[31 + n*(len(tlmy_struct_iface)) + 3 ], 
                                                            tmp[31 + n*(len(tlmy_struct_iface)) + 4 ],
                                                            tmp[31 + n*(len(tlmy_struct_iface)) + 5 ],
                                                            tmp[31 + n*(len(tlmy_struct_iface)) + 6 ],
                                                            tmp[31 + n*(len(tlmy_struct_iface)) + 7 ],
                                                            tmp[31 + n*(len(tlmy_struct_iface)) + 10])

			
                        tlmytlmy.append('%.03f %d %0.02f %d %d %d %d %s\n' % (pk.getSecHdrTFieldC() + pk.getSecHdrTFieldF()*(10**-3), pk.getPacketSequenceCount(), tmp[10], tmp[11], tmp[23], tmp[24], tmp[26], s))
                        avgload.append(tmp[10])
                        if firstmem == 0:
                            firstmem = tmp[11]

                    

        #### Operaciones con paquetes
        if options.packet:
            pk = SpacePacket()
            
            # Parsear cabecera
            pk.parseHeader(fd.read(6))
            
            packetcount = packetcount + 1

            # Obtener valores
            pkcounter = pk.getPacketSequenceCount()
            size      = pk.getPacketDataLength() + 1
            apid      = pk.getAPID()

            # Ya se que apid estoy procesando, creo los directorios y archivos:
            if firstpacket == None:
                fdout = open(options.filename+'-Reporte_Packet.txt', 'w')
                if options.valverif >= 0:
                    fdvalverif = open(options.filename+'-ByteVerif.txt', 'w')


            # Guardo primer contador
            if firstpacket == None:
                firstpacket = (packetcount, pkcounter)

            # Guardo el ultimo
            lastpacket = (packetcount, pkcounter)

            # Calculo de paquetes perdidos y reporte
            if packetant != 0 and firstpacket != None:
                diff = pkcounter - packetant
                if diff != 1 and diff > 0:
                    packetloss = packetloss + diff - 1
                    print 'WARN: Packet lost - prev: %d, curr: %d' % (packetant, pkcounter)
            packetant = pkcounter

            # Avanzo y seteo el payload
            pk.setPayload(fd.read(size))

            # Verificacion a nivel Byte
            if options.valverif >= 0:
                tot   = 0
                for byte in pk.getPayload():
                    if ord(byte) != options.valverif:
                        tot = tot + 1
                    totbytes = totbytes + 1

                valverif[pkcounter] = tot
                totvalverif = totvalverif + tot


    print
    print 'Operacion finalizada...'
    print 

    if options.frame:
        print 'Reporte generado: %s\n\n' % fdout.name
    if options.packet:
        print 'Reporte generado: %s\n\n' % fdout.name
    if options.valverif >= 0:
        print 'Listado generado: %s\n\n' % fdvalverif.name


    # GENERAR REPORTES
    # General
    fdout.write('***************\n')
    fdout.write('*** Reporte ***\n')
    fdout.write('***************\n')
    fdout.write('\n')
    fdout.write('Archivo de entrada: %s\n\n' % options.filename)
    fdout.write('Linea de comando: ')
    for arg in sys.argv:
        fdout.write('%s ' % arg)
    fdout.write('\n\n')


    # Reporte para modo frame
    if options.frame and options.counters:
        fdout.write('Frames procesados: %d\n' % lastframe[0])
        fdout.write('Frames perdidos  : %d\n' % frameloss)
        fdout.write('Porcentaje       : %f %%\n\n' % ((lastframe[0] * 100.)/(lastframe[0]+frameloss)))

        fdout.write('\tFrame Count\tCCSDS Frame Counter\n')
        fdout.write('First:\t%d\t\t%d\n' % (firstframe[0], firstframe[1]))
        fdout.write('Last:\t%d\t\t%d\n\n'  % (lastframe[0], lastframe[1]))

        fdout.write('APIDs encontrados (APID - cantidad - tamanno): \n')
        for apid in sorted(apidcount.iterkeys()):
            fdout.write('0x%03X - %d\t- %d\n' % (apid, apidcount[apid], apidsizes[apid]))
        fdout.write('\n')

        #fdout.write('Paquetes con Sec. HDR: %d' % pksechdrcount)
        #fdout.write('\n')

        fdout.write('Cadencia promedio de los paquetes:\n')

        for apid in sorted(tapid.iterkeys()):
            s = 0
            for i in range(0, len(tapid[apid][1]) - 2):
                s += tapid[apid][1][i+1] - tapid[apid][1][i]
            try:
                fdout.write('0x%03X\t - %.02f\n' % (apid, 1/(s/tapid[apid][0])))
            except:
                fdout.write('0x%03X\t - no valido\n' % (apid))

        fdout.write('\n')
        fdout.write('APIDs no reportados (apid, offset en el archivo):\n')

        if len(badapidoffset.keys()) == 0:
            fdout.write('-\n')
        else:
            for apid in sorted(badapidoffset.iterkeys()):
                fdout.write('0x%03X - 0x%X\n' % (apid, badapidoffset[apid]))
            fdout.write('\n')

        # Carga de sistema promedio de la computadora de TLMY
        carga = 0
        for load in avgload:
            carga = carga + load
        carga = carga / len(avgload)
        
        lastmem = tmp[11]

        fdout.write('\nEstado del Sistema\n')
        fdout.write('==================\n')
        fdout.write('\nCarga Promedio: %f\n' % carga)
        fdout.write('Memoria Libre inicial (MB): %d\n' % (firstmem/1024/1024))
        fdout.write('Memoria Libre final   (MB): %d\n' % (lastmem/1024/1024))


        # Tabla de los subsistemas
        fdout.write('\n\n')
        fdout.write('Tabla de Subsistemas\n')
        fdout.write('====================\n\n')
        
        fdout.write('Subsys\t\tAPIDQ\tPR\t\tUSD\tIC4PS\t\tTimeout\n')
        fdout.write('=======================================================================\n')
        for n in range(0,N_ROUTERS):
            fdout.write('%s\t\t%d\t%d\t\t%d\t%d\t\t%d\n' % (iface2txt[tmp[31 + n*(len(tlmy_struct_iface)) + 3 ]],
                                                            tmp[31 + n*(len(tlmy_struct_iface)) + 4 ],
                                                            tmp[31 + n*(len(tlmy_struct_iface)) + 5 ],
                                                            tmp[31 + n*(len(tlmy_struct_iface)) + 6 ],
                                                            tmp[31 + n*(len(tlmy_struct_iface)) + 7 ],
                                                            tmp[31 + n*(len(tlmy_struct_iface)) + 10 ]
                                                            ))
        
    # Reporte para modo packet
    if options.packet and (options.counters or options.valverif >= 0):
        fdout.write ('Packets procesados: %d\n' % lastpacket[0])
        fdout.write ('Packets perdidos  : %d\n' % packetloss)
        fdout.write ('Porcentaje        : %f %%\n\n' % ((lastpacket[0] * 100.)/(lastpacket[0]+packetloss)))

#((packetloss*100.)/(lastpacket[0]+packetloss)))

        fdout.write('\tPacket Count\tCCSDS Packet Counter\n')
        fdout.write('First:\t%d\t\t%d\n' % (firstpacket[0], firstpacket[1]))
        fdout.write('Last:\t%d\t\t%d\n'  % (lastpacket[0], lastpacket[1]))

        if options.valverif >= 0:
            fdout.write('\nVerificacion por valor:\n')
            fdout.write('Chequeado contra: 0x%02X\n' % options.valverif)
            fdout.write('Bytes Totales   : %d\n' % totbytes)
            fdout.write('Bytes Erroneos  : %d\n' % totvalverif)
            fdout.write('Porcentaje      : %f %%\n' % ((totvalverif*100.)/totbytes))
            
    
    # Cierro reporte
    fdout.write('\n\n')
    fdout.write('*******************\n')
    fdout.write('*** Fin Reporte ***\n')
    fdout.write('*******************\n')

    if options.frame:
        print 'Creando archivos de tabla de paquetes...'
        for d in tpackets:
            fdout2.write(d)
        if options.tlmy:
            print 'Creando archivo de tablas de tlmy tlmy...'
            for d in tlmytlmy:
                fdout3.write(d)

    if options.valverif >= 0:
        for key in valverif.keys():
            fdvalverif.write('%d %d\n' % (key, valverif[key]))
        fdvalverif.close()
                             
                             
    # Aqui se cierran todos los archivos
    fd.close()
    fdout.close()
    if options.frame:
        fdout2.close()
        if options.tlmy:
            fdout3.close()

    if options.valverif >= 0:
        fdvalverif.close()
    

if __name__ == "__main__":
    sys.exit(main())
